from typing import List, Any, Union, Optional
import numbers
import numpy as np
import torch
# from icecream import ic


from .cayley import get_cayley_tensor, blades_from_bases
from .blades import (
    BladeKind, get_blade_of_kind_indices, get_blade_indices_from_names,
    get_blade_repr, invert_blade_indices
)
from .mv_ops import mv_multiply, mv_reversion, mv_grade_automorphism, mv_conv1d, f_mv_conv1d, mv_multiply_element_wise
from .mv import MultiVector
from .torch_ga import GeometricAlgebra
from .jacobian import get_jacobian



def compute_rotation_plane(points, referencepoint):   
    p0, p1 =  referencepoint , points
    lines = p0 & p1
    midpoints = 0.5*(p0 + p1)
    rotationplane = lines | midpoints 
    return rotationplane

def compute_log_det_jacobian_transformation(ga,points,referencepoint):
    j2 = get_jacobian(lambda p1: compute_rotation_plane(ga(p1),referencepoint).tensor , points.tensor)
    J = j2.j[...,slice(1,5), slice(11,15)]
    return torch.log(torch.linalg.det(J).abs())



"""
Utility functions to converty from vector to GA representation. they assume G(0,1,1,1)

"""


from typing import List, Any, Union, Optional
import numbers
import numpy as np
import torch
# from icecream import ic


from .cayley import get_cayley_tensor, blades_from_bases
from .blades import (
    BladeKind, get_blade_of_kind_indices, get_blade_indices_from_names,
    get_blade_repr, invert_blade_indices
)
from .mv_ops import mv_multiply, mv_reversion, mv_grade_automorphism, mv_conv1d, f_mv_conv1d, mv_multiply_element_wise
from .mv import MultiVector
from .torch_ga import GeometricAlgebra


# Probably the primary encoding
# 

def plane2pga(ga, distance, normal_dir):
    """
    Plane w/ normal direction n in R3, origin distance d in R
    """
    inputs = torch.cat([distance,normal_dir],dim=-1)
    blade_indices = (ga.blade_degrees==1).long()
    blade_indices = torch.where(blade_indices)[0]
    tensor = ga.from_tensor(inputs, blade_indices)
    return tensor

def pga2plane(ga,tensor):
    if isinstance(tensor, MultiVector): tensor = tensor.tensor
    blade_indices = (ga.blade_degrees==1).long()
    blade_indices = torch.where(blade_indices)[0]
    dn = tensor[...,blade_indices]
    distance,normal_dir = dn[...,:1],dn[...,1:]
    return distance,normal_dir

def line2pga(ga, orthogonal_shift, direction):
    """
    Line w/ direction n in R3, orthogonal shift s in R3
    """    
    inputs = torch.cat([orthogonal_shift,direction],dim=-1)
    blade_indices = (ga.blade_degrees==ga.p-1).long()
    blade_indices = torch.where(blade_indices)[0]
    tensor = ga.from_tensor(inputs, blade_indices)
    return tensor

def pga2line_(ga,tensor):
    if isinstance(tensor, MultiVector): tensor = tensor.tensor
    blade_indices = (ga.blade_degrees==ga.p-1).long()
    blade_indices = torch.where(blade_indices)[0]
    # ic(blade_indices,tensor.shape)
    od = tensor[...,blade_indices]
    orthogonal_shift, direction = od[...,:ga.p],od[...,ga.p:]
    return orthogonal_shift, direction

def pga2line(ga,tensor):
    if not isinstance(tensor, MultiVector): 
        lines = ga(tensor)
    else:
        lines = tensor
    
    ORIG = ga(ga.e123 if ga.p==3 else ga.e12)
    d = (lines|ORIG)*lines
    points = pga2point(ga,d)
    # ic(points)
    orthogonal_shift, direction = pga2line_(ga,tensor)
    return points, direction
    
    
    # if isinstance(tensor, MultiVector): tensor = tensor.tensor
    # blade_indices = (ga.blade_degrees==2).long()
    # blade_indices = torch.where(blade_indices)[0]
    # od = tensor[...,blade_indices]
    # orthogonal_shift, direction = od[...,:3],od[...,3:]
    # return orthogonal_shift, direction

def point2pga(ga, point):
    """
    Point p in R3
    """    
    inputs = torch.cat([point,torch.ones([*point.shape[:-1],1]).to(point)],dim=-1)
    blade_indices = (ga.blade_degrees==ga.p).long()
    blade_indices = torch.where(blade_indices)[0]
    tensor = ga.from_tensor(inputs, blade_indices)
    # /inputs.tensor[...,-2]
    return tensor

def pga2point(ga,tensor):
    if isinstance(tensor, MultiVector): tensor = tensor.tensor
    blade_indices = (ga.blade_degrees==ga.p).long()
    blade_indices = torch.where(blade_indices)[0]
    od = tensor[...,blade_indices]
    point, one_ = od[...,:ga.p],od[...,ga.p:]
    if True: point = point/one_
    return point

# CGA

def point2cga(ga, point):
    """
    Point p in R3
    """    
    inputs = torch.cat([point,torch.ones([*point.shape[:-1],1]).to(point)],dim=-1)
    blade_indices = torch.tensor([5,12,14,15])
    tensor = ga.from_tensor(inputs, blade_indices)
    # /inputs.tensor[...,-2]
    return tensor

def cga2point(ga,tensor):
    if isinstance(tensor, MultiVector): tensor = tensor.tensor
    blade_indices = torch.tensor([5,12,14,15])
    od = tensor[...,blade_indices]
    point, one_ = od[...,:3],od[...,3:]
    if True: point = point/one_
    return point



# Direction
# 415,425,435,
# -21,-24,-25

# Moment
# 235,315,125
# 23,-20,18

def line2cga(ga, moment, direction):
    """
    Line w/ direction n in R3, orthogonal shift s in R3
    """    
    scaling = torch.tensor([1,-1,1]).to(moment)
    inputs = torch.cat([moment*scaling,-direction],dim=-1)
    blade_indices = torch.tensor([23,20,18,21,24,25])
    tensor = ga.from_tensor(inputs, blade_indices)
    return tensor

def cga2line_(ga,tensor):
    if isinstance(tensor, MultiVector): tensor = tensor.tensor
    blade_indices = torch.tensor([23,20,18,21,24,25])
    od = tensor[...,blade_indices]
    moment, direction = od[...,:3],-od[...,3:]
    scaling = torch.tensor([1,-1,1]).to(moment)
    return moment*scaling, direction

def cga2line(ga,tensor):
    if not isinstance(tensor, MultiVector): 
        lines = ga(tensor)
    else:
        lines = tensor
    
    ORIG = ga(ga.e3 - ga.e4)
    d = (lines|ORIG)*lines
    points = cga2point(ga,d)
    moment, direction = cga2line_(ga,tensor)
    return points, direction
    
    

# Dual encoding
# 

def dual_plane2pga(ga, distance, normal):
    n = ga(ga.from_tensor(normal,torch.tensor([2,3,4])))
    delta = ga(ga.from_tensor(distance,torch.tensor([0])))
    tensor = n - delta*ga(ga.e0)
    # tensor = n - ga(distance*ga.e0)
    return tensor.tensor

def dual_pga2plane(ga,tensor):
    if isinstance(tensor, MultiVector): tensor = tensor.tensor
    normal, distance = tensor[...,2:5],-tensor[...,1]
    return distance, normal

def dual_line2pga(ga, orthogonal_shift, direction):
    """
    Line w/ direction n in R3, orthogonal shift s in R3
    """
    n1n2 = ga(ga.from_tensor(direction,torch.tensor([2,3,4])))
    distance = ga(ga.from_tensor(orthogonal_shift,torch.tensor([2,3,4])))
    tensor = n1n2 - ga(ga.e0)*distance*n1n2
    return tensor.tensor

def dual_pga2line(ga,tensor):
    if isinstance(tensor, MultiVector): tensor = tensor.tensor
    n1n2 = ga(ga.from_tensor(tensor[...,2:5],torch.tensor([2,3,4])))
    tensor = ga(tensor) - n1n2  
    gad = GeometricAlgebra([1]*ga.p+[-1]*ga.q)
    dn1n2 = gad(gad.from_tensor(tensor.tensor[...,[1,-5,-4,-3]],torch.tensor([0,4,5,6])))
    n1n2 = gad(gad.from_tensor(n1n2.tensor[...,[2,3,4]],torch.tensor([1,2,3])))
    distance = dn1n2/n1n2
    orthogonal_shift, direction = -distance.tensor[...,[1,2,3]], n1n2.tensor[...,[1,2,3]]
    return orthogonal_shift, direction

def dual_point2pga(ga, point):
    """
    Point p in R3
    """    
    x = ga(ga.from_tensor(point,torch.tensor([2,3,4])))
    tensor = (ga(ga.e_)-ga(ga.e0)*x)*ga(ga.e123)
    return tensor.tensor

def dual_pga2point(ga,tensor):
    if isinstance(tensor, MultiVector): tensor = tensor.tensor
    point = ga(tensor).dual()
    tensor = point.tensor[...,2:5]
    return tensor


# Need to be verified!!!
# 
# 

# Operations
def reflection2pga(ga, distance, normal_dir):
    """
    Reflection through plane w/ normal n in R3, origin shift d in R
    """
    inputs = torch.cat([distance,normal_dir],dim=-1)
    blade_indices = (ga.blade_degrees==1).long()
    blade_indices = torch.where(blade_indices)[0]
    tensor = ga.from_tensor(inputs, blade_indices)
    return tensor

def pga2reflection(ga,tensor):
    if isinstance(tensor, MultiVector): tensor = tensor.tensor
    blade_indices = (ga.blade_degrees==1).long()
    blade_indices = torch.where(blade_indices)[0]
    dn = tensor[...,blade_indices]
    distance,normal_dir = dn[...,:1],dn[...,1:]
    return distance,normal_dir

# translation 
def translation2pga(ga, translation):
    """
    Translation t in R3
    """    
    _one = torch.ones([*translation.shape[:-1],1]).to(translation)
    inputs = torch.cat([_one,translation*0.5],dim=-1)
    blade_indices = (ga.blade_degrees==2).long() 
    blade_indices = torch.where(blade_indices)[0][:3]
    tensor = ga.from_tensor(inputs, [0]+blade_indices)
    return tensor

def pga2translation(ga,tensor):
    if isinstance(tensor, MultiVector): tensor = tensor.tensor
    blade_indices = (ga.blade_degrees==2).long()
    blade_indices = torch.where(blade_indices)[0]
    od = tensor[...,blade_indices[:3]]
    _one, translation = od[...,:3],od[...,3:]
    return translation*2

# rotation 
def rotation2pga(ga, rotation):
    """
    Translation t in R3
    """    
    
    inputs = rotation
    blade_indices = (ga.blade_degrees==2).long() 
    blade_indices = torch.where(blade_indices)[0][3:]
    tensor = ga.from_tensor(inputs, [0]+blade_indices)
    return tensor

def pga2rotation(ga,tensor):
    if isinstance(tensor, MultiVector): tensor = tensor.tensor
    blade_indices = (ga.blade_degrees==2).long()
    blade_indices = torch.where(blade_indices)[0]
    rotation = tensor[...,[0]+blade_indices[3:]]
    return rotation

# Point reflection
def pointreflection2pga(ga, point):
    """
    Point reflection through p in R3
    """    
    inputs = torch.cat([point,torch.ones([*point.shape[:-1],1]).to(point)],dim=-1)
    blade_indices = (ga.blade_degrees==3).long()
    blade_indices = torch.where(blade_indices)[0]
    tensor = ga.from_tensor(inputs, blade_indices)
    return tensor

def pga2pointreflection(ga,tensor):
    if isinstance(tensor, MultiVector): tensor = tensor.tensor
    blade_indices = (ga.blade_degrees==3).long()
    blade_indices = torch.where(blade_indices)[0]
    od = tensor[...,blade_indices]
    point, one_ = od[...,:3],od[...,3:]
    if False: point = point/one_
    return point




"""Defines the `MultiVector` class which is used as a convenience wrapper
for `GeometricAlgebra` operations.
"""
from __future__ import annotations
from typing import Union

from .blades import BladeKind

import torch
# from .torch_ga import GeometricAlgebra
import torch_ga

# from icecream import ic
class MultiVector:
    """Wrapper for geometric algebra tensors using `GeometricAlgebra`
    operations in a less verbose way using operators.
    """

    def __init__(self, blade_values: torch.Tensor, algebra: torch_ga.GeometricAlgebra):
        """Initializes a MultiVector from a geometric algebra `torch.Tensor`
        and its corresponding `GeometricAlgebra`.

        Args:
            blade_values: Geometric algebra `torch.Tensor` with as many elements
            on its last axis as blades in the algebra
            algebra: `GeometricAlgebra` instance corresponding to the geometric
            algebra tensor
        """

        self._blade_values = blade_values
        self._algebra = algebra

    def split(self):
        """blades multivector."""
        return self._algebra.split(self._blade_values)

    @property
    def blades(self):
        """blades multivector."""
        return self._algebra.blades

    @property
    def blades_numbers(self):
        """blades multivector."""
        return self._algebra.blades_numbers

    @property
    def scalar(self):
        """scalar part of multivector."""
        return self._blade_values[...,0]
    @property
    def pseudo_scalar(self):
        """scalar part of multivector."""
        return self._blade_values[...,-1]
    @property
    def vector(self):
        """vector part of multivector."""
        return self._blade_values[...,1:self.algebra.dim+1]
    @property
    def pseudo_vector(self):
        """vector part of multivector."""
        return self._blade_values[...,-self.algebra.dim-1:-1]

    @property
    def bivector(self):
        """bivector part of multivector."""        
        return self._blade_values[...,self.algebra.blade_degrees==2]
    @property
    def pseudo_bivector(self):
        """pseudo bivector part of multivector."""
        return self._blade_values[...,self.algebra.blade_degrees==(self.algebra.dim-2)]

    @property
    def trivector(self):
        """bivector part of multivector."""        
        return self._blade_values[...,self.algebra.blade_degrees==3]
    @property
    def pseudo_trivector(self):
        """pseudo bivector part of multivector."""
        return self._blade_values[...,self.algebra.blade_degrees==(self.algebra.dim-3)]
        
    @property
    def tensor(self):
        """Geometric algebra tensor holding the values of this multivector."""
        return self._blade_values

    @property
    def algebra(self):
        """`GeometricAlgebra` instance this multivector belongs to."""
        return self._algebra

    @property
    def batch_shape(self):
        """Batch shape of the multivector (ie. the shape of all axes except
        for the last one in the geometric algebra tensor).
        """
        return self._blade_values.shape[:-1]

    def __len__(self) -> int:
        """Number of elements on the first axis of the geometric algebra
        tensor."""
        return self._blade_values.shape[0]

    def __iter__(self):
        for n in range(self._blade_values.shape[0]):
            # If we only have one axis left, return the
            # actual numbers, otherwise return a new
            # multivector.
            if self._blade_values.shape.ndims == 1:
                yield self._blade_values[n]
            else:
                yield MultiVector(
                    self._blade_values[n],
                    self._algebra
                )

    def __lshift__(self, other: MultiVector) -> MultiVector:
        """
        Left contraction
        """
        assert isinstance(other, MultiVector), f"{type(other)}"
        return MultiVector(
            self._algebra.right_contraction(self._blade_values, other._blade_values),
            self._algebra
        )
    def __rshift__(self, other: MultiVector) -> MultiVector:
        """
        Right contraction
        """
        assert isinstance(other, MultiVector), f"{type(other)}"
        return MultiVector(
            self._algebra.right_contraction(self._blade_values, other._blade_values),
            self._algebra
        )
    

    def __xor__(self, other: MultiVector) -> MultiVector:
        """Exterior product. See `GeometricAlgebra.ext_prod()`"""
        assert isinstance(other, MultiVector), f"{type(other)}"

        return MultiVector(
            self._algebra.ext_prod(self._blade_values, other._blade_values),
            self._algebra
        )

    def __or__(self, other: MultiVector) -> MultiVector:
        """Inner product. See `GeometricAlgebra.inner_prod()`"""
        assert isinstance(other, MultiVector), f"{type(other)}"

        # return MultiVector(
        #     self._algebra.inner_prod(self._blade_values, other._blade_values),
        #     self._algebra
        # )
        return MultiVector(self._algebra.inner_prod(self._blade_values, other._blade_values),self._algebra)

    def __mul__(self, other: MultiVector) -> MultiVector:
        """Geometric product. See `GeometricAlgebra.geom_prod()`"""
        
        if type(other) in (int, float, torch.Tensor):
            return MultiVector(self._blade_values*other,self._algebra)
        
        assert isinstance(other, MultiVector), f"{type(other)}"

        return MultiVector(self._algebra.geom_prod(self._blade_values, other._blade_values),self._algebra)
    __rmul__=__mul__
    
    def __truediv__(self, other: MultiVector) -> MultiVector:
        """Division, ie. multiplication with the inverse."""
        if isinstance(other,(int,float,torch.Tensor)):
            return MultiVector(
                self._blade_values/other,                
                self._algebra
            )
        
        assert isinstance(other, MultiVector), f"{type(other)}"
        return MultiVector(
            self._algebra.geom_prod(
                self._blade_values,
                self._algebra.inverse(other._blade_values)
            ),
            self._algebra
        )

    def __and__(self, other: MultiVector) -> MultiVector:
        """Regressive product. See `GeometricAlgebra.reg_prod()`"""        
        assert isinstance(other, MultiVector), f"{type(other)}"                
        return MultiVector(
            self._algebra.reg_prod(self._blade_values, other._blade_values),
            self._algebra
        )

    def __invert__(self) -> MultiVector:
        """Reversion. See `GeometricAlgebra.reversion()`"""
        return MultiVector(
            self._algebra.reversion(self._blade_values),
            self._algebra
        )

    def __neg__(self) -> MultiVector:
        """Negation."""
        return MultiVector(
            -self._blade_values,
            self._algebra
        )

    def __add__(self, other: MultiVector) -> MultiVector:
        """Addition of multivectors."""
        if type(other) in (int, float):
            _tmp = self._blade_values.clone()
            _tmp[...,0]+=other
            return MultiVector(_tmp,self._algebra)        
        
        assert isinstance(other, MultiVector), f"{type(other)}"
        
        return MultiVector(
            self._blade_values + other._blade_values,
            self._algebra
        )
    __radd__=__add__
    
    def __sub__(self, other: MultiVector) -> MultiVector:
        """Subtraction of multivectors."""
        if type(other) in (int, float):
            _tmp = self._blade_values.clone()
            _tmp[...,0]-=other
            return MultiVector(_tmp,self._algebra)        
        
        assert isinstance(other, MultiVector), f"{type(other)}"

        return MultiVector(
            self._blade_values - other._blade_values,
            self._algebra
        )
    def __rsub__(a,b):
        return b + -1 * a        

    def __pow__(self, n: int) -> MultiVector:
        """Multivector raised to an integer power."""
        return MultiVector(
            self._algebra.int_pow(self._blade_values, n),
            self._algebra
        )

    def __getitem__(self, key: Union[str, list[str]]) -> MultiVector:
        """`MultiVector` with only passed blade names as non-zeros."""
        if type(key) in [int, float, slice]:
            return MultiVector(
                self._blade_values[key],
                self._algebra
            )
            
        return MultiVector(
            self._algebra.keep_blades_with_name(self._blade_values, key),
            self._algebra
        )

    def __call__(self, key: Union[str, list[str]]):
        """`torch.Tensor` with passed blade names on last axis."""
        return self._algebra.select_blades_with_name(self._blade_values, key)

    def __repr__(self) -> str:
        return self._algebra.mv_repr(self._blade_values)

    def proj(self, other: MultiVector) -> MultiVector:
        """meet of multivectors."""
        assert isinstance(other, MultiVector), f"{type(other)}"
        return MultiVector(
            self._algebra.projection(other._blade_values, self._blade_values),
            self._algebra
        )
    def proj_on(self, other: MultiVector) -> MultiVector:
        """meet of multivectors."""
        assert isinstance(other, MultiVector), f"{type(other)}"
        return MultiVector(
            self._algebra.projection(self._blade_values, other._blade_values),
            self._algebra
        )

    def sandwitch(self, other: MultiVector) -> MultiVector:
        """geometric product of b a b~, b is the other, a is myself"""
        assert isinstance(other, MultiVector), f"{type(other)}"
        return MultiVector(
            self._algebra.sandwitch_prod(self._blade_values, other._blade_values),
            self._algebra
        )

    def meet(self, other: MultiVector) -> MultiVector:
        """meet of multivectors."""
        assert isinstance(other, MultiVector), f"{type(other)}"
        return MultiVector(
            self._algebra.meet(self._blade_values, other._blade_values),
            self._algebra
        )
    def join(self, other: MultiVector) -> MultiVector:
        """meet of multivectors."""
        assert isinstance(other, MultiVector), f"{type(other)}"
        return MultiVector(
            self._algebra.join(self._blade_values, other._blade_values),
            self._algebra
        )

    def inverse(self) -> MultiVector:
        """Inverse. See `GeometricAlgebra.inverse()`."""
        return MultiVector(
            self._algebra.inverse(self._blade_values),
            self._algebra
        )
        
    def inv(self) -> MultiVector:
        """Inverse. See `GeometricAlgebra.inverse()`."""
        return self.inverse()
    
    @property
    def shape(self):
        return self._blade_values.shape[:-1]

    def simple_inverse(self) -> MultiVector:
        """Simple inverse. See `GeometricAlgebra.simple_inverse()`."""
        return MultiVector(
            self._algebra.simple_inverse(self._blade_values),
            self._algebra
        )

    def dual(self) -> MultiVector:
        """Dual. See `GeometricAlgebra.dual()`."""
        return MultiVector(
            self._algebra.dual(self._blade_values),
            self._algebra
        )
    def grade(self,_grade=None):
        """
        Retruns the blade of the specified grade.
        """
        if _grade is None:
            return self._algebra.grade(self._blade_values)
        
        return MultiVector(
            self._algebra.grade(self._blade_values,_grade),
            self._algebra
        )
    # def grade(self):
    #     """
    #     Retruns the max grade of all non zeros blades, if zero, return zero. 
    #     """
    #     return self._algebra.grade(self._blade_values)

    def norm(self):
        return self._algebra.norm(self._blade_values)
    def inorm(self):
        return self._algebra.inorm(self._blade_values)
    def normalized(self):
        # return self._algebra.normalized(self._blade_values)  
        return MultiVector(
            self._algebra.normalized(self._blade_values),
            self._algebra
        )


    def conjugation(self) -> MultiVector:
        """Conjugation. See `GeometricAlgebra.conjugation()`."""
        return MultiVector(
            self._algebra.conjugation(self._blade_values),
            self._algebra
        )
        
    def reversion(self) -> MultiVector:
        """reversion. See `GeometricAlgebra.reversion()`."""
        return MultiVector(
            self._algebra.reversion(self._blade_values),
            self._algebra
        )

    def grade_automorphism(self) -> MultiVector:
        """Grade automorphism. See `GeometricAlgebra.grade_automorphism()`."""
        return MultiVector(
            self._algebra.grade_automorphism(self._blade_values),
            self._algebra
        )

    def approx_exp(self, order: int = 50) -> MultiVector:
        """Approximate exponential. See `GeometricAlgebra.approx_exp()`."""
        return MultiVector(
            self._algebra.approx_exp(self._blade_values, order=order),
            self._algebra
        )

  
    def exp(self, square_scalar_tolerance: Union[float, None] = 1e-4) -> MultiVector:
        """Exponential. See `GeometricAlgebra.exp()`."""
        return MultiVector(
            self._algebra.exp(
                self._blade_values,
                square_scalar_tolerance=square_scalar_tolerance
            ),
            self._algebra
        )

    def approx_log(self, order: int = 50) -> MultiVector:
        """Approximate logarithm. See `GeometricAlgebra.approx_log()`."""
        return MultiVector(
            self._algebra.approx_log(self._blade_values, order=order),
            self._algebra
        )

    def is_pure_kind(self, kind: BladeKind) -> bool:
        """Whether the `MultiVector` is of a pure kind."""
        return self._algebra.is_pure_kind(self._blade_values, kind=kind)

    def geom_conv1d(self, kernel: MultiVector,
                    stride: int, padding: str,
                    dilations: Union[int, None] = None) -> MultiVector:
        """1D convolution. See `GeometricAlgebra.geom_conv1d().`"""
        return MultiVector(
            self._algebra.geom_conv1d(
                self._blade_values, kernel._blade_values,
                stride=stride, padding=padding, dilations=dilations
            ),
            self._algebra
        )
    def __pow__(a,b):
        assert(isinstance(b,int) and b >= 0), "power only defined on integers"
        if b==0: return a._algebra.e_
        tmp = a
        for ki in range(b-1):
           tmp = tmp*a
        return tmp



import torch

import einops
from collections import namedtuple
from typing import Iterable


__all__ = [
    "get_jacobian"
]

Jacobian = namedtuple("Jacobian", ["y", "j"])

from functools import partial


def get_jacobian(fun, x, m=None, slice_in=None, slice_out=None):
    """
    Computes the jacobian of `fun`-tion with respect to the variable `x`.
    Batch evaluates the jacobian with cost of O(1) vs O(d), but has memory cost of O(d) (only work for low d, d <~ 1000).
    Parameters:
    -----------
    fun: callable: function for the jacobian
    x: torch.Tensor : coordinated of the jacobian
    slice_in, slice_out: if not None, will slice the input or output of the Jacobian

    Returns:
    --------
    jac: Jacobian
        Named tuple containing members `y` and `j`.
        `y` is the function value  of `fun` evaluated at `x`
        `j` is the jacobian matrix of `fun` evaluated at `x`
        
    """
    if len(x.shape)==1: x.unsqueeze(0)
    shape = x.shape[:-1]
    d = x.shape[-1]
    x = x.view(-1, d)
    n = x.shape[0]
    z = einops.repeat(x, "n j -> (n i) j", i=d)
    z.requires_grad_(True)
    y = fun(z)    
    if  m is None:
        out_grad = torch.eye(d, device=x.device, dtype=x.dtype).tile(n, 1)
    else:        
        out_grad = torch.zeros(d, m, device=x.device, dtype=x.dtype).tile(n, 1)        
        out_grad[slice_out,slice_in] = 1.
    j = torch.autograd.grad(y, z, out_grad, create_graph=True, retain_graph=True)[0].view(*shape, d, d)

    if not slice_in is None:
        j = j[...,slice_out,slice_in]
    
    return Jacobian(
        y=einops.rearrange(y, "(n i) j -> n i j", i=d)[:, 0, :].view(*shape, -1),
        j=j
    )




"""
PyTorch Geometric Algebra

An python library for Geometric Algebra in Pytorch
"""

__author__ = 'Francesco Alesiani'
__credits__ = 'Torch-GA'

# from .torch_ga import GeometricAlgebra
from .torch_ga import *
from .blades import *
from .cayley import *
from .layers import *
from .mv_ops import *
from .mv import *
from .utils import *
from .plots import *
from .jacobian import *
from .transform import *



"""
Utility functions to plot lines, planes and points in 3D. they assume G(0,1,1,1)

"""

import numpy as np
# import meshplot as mp
import matplotlib.pyplot as plt
# from icecream import ic

from .utils import pga2line, pga2plane, pga2point, line2pga, plane2pga, point2pga
from .utils import cga2point, point2cga, cga2line

from .mv import MultiVector
# from .torch_ga import GeometricAlgebra
import torch

def t2np(x):
    if isinstance(x,torch.Tensor):
        return x.detach().cpu().numpy()
    return x

def plot_line_pga(ga,tensor,ax=None,alpha = 1.0, **kargs):
    if isinstance(tensor, MultiVector): tensor = tensor.tensor
    if len(tensor.shape)==1: tensor=tensor.unsqueeze(0)
    if ax is None:
        fig = plt.figure(figsize=(12, 12))
        if ga.p==3: ax = fig.add_subplot(projection='3d')   
        else: ax = fig.add_subplot()   
            
    points, direction = pga2line(ga,tensor)    
    # ic(points,direction)
    # orthogonal_shift, direction = plucker_to_line(orthogonal_shift, direction)
    for s,d in zip(t2np(points), t2np(direction)):        
        plot_line_xyz(s,d,ax, ga.p, alpha, **kargs) 
        
    if True:
        coords = t2np(points)
        for ki,xy in enumerate(coords):
            ax.text(*xy,f"{ki}")        

def plot_plane_pga(ga,tensor,ax=None, **kargs):
    if isinstance(tensor, MultiVector): tensor = tensor.tensor
    if len(tensor.shape)==1: tensor=tensor.unsqueeze(0)
    if ax is None:
        fig = plt.figure(figsize=(12, 12))
        # ax = fig.add_subplot(projection='3d')    
        if ga.p==3: ax = fig.add_subplot(projection='3d')   
        else: ax = fig.add_subplot()   
        
    points, direction = pga2plane(ga,tensor)
    if ga.p==3:
        xx, yy = np.meshgrid(np.linspace(-1,1,10),np.linspace(-1,1,10))
    else:
        xx = np.linspace(-1,1,10)
    for d,n in zip(t2np(points), t2np(direction)):
        if ga.p==3: plot_plane_xyz(d,n,ax,xx,yy,**kargs)
        if ga.p==2: plot_plane_xy(d,n,ax,xx,**kargs)
        
    # coords = t2np(direction)
    # for ki,xy in enumerate(coords):
    #     ax.text(*xy,f"{ki}")              

def plot_point_pga(ga,tensor, ax=None, **kargs):
    if isinstance(tensor, MultiVector): tensor = tensor.tensor
    if len(tensor.shape)==1: tensor=tensor.unsqueeze(0)
    if ax is None:
        fig = plt.figure(figsize=(12, 12))
        # ax = fig.add_subplot(projection='3d')    
        if ga.p==3: ax = fig.add_subplot(projection='3d')   
        else: ax = fig.add_subplot()   
        
    point = pga2point(ga,tensor)
    ax.scatter(*t2np(point).T,**kargs)
    
    coords = t2np(point)
    for ki,xy in enumerate(coords):
        ax.text(*xy,f"{ki}")

# CGA

def plot_point_cga(ga,tensor, ax=None, **kargs):
    if isinstance(tensor, MultiVector): tensor = tensor.tensor
    if len(tensor.shape)==1: tensor=tensor.unsqueeze(0)
    if ax is None:
        fig = plt.figure(figsize=(12, 12))
        # ax = fig.add_subplot(projection='3d')    
        if ga.p==3: ax = fig.add_subplot(projection='3d')   
        else: ax = fig.add_subplot()   
        
    point = cga2point(ga,tensor)
    ax.scatter(*t2np(point).T,**kargs)
    
    coords = t2np(point)
    for ki,xy in enumerate(coords):
        ax.text(*xy,f"{ki}")


def plot_line_cga(ga,tensor,ax=None,alpha = 1.0, **kargs):
    if isinstance(tensor, MultiVector): tensor = tensor.tensor
    if len(tensor.shape)==1: tensor=tensor.unsqueeze(0)
    if ax is None:
        fig = plt.figure(figsize=(12, 12))
        # ax = fig.add_subplot(projection='3d')    
        if ga.p==3: ax = fig.add_subplot(projection='3d')   
        else: ax = fig.add_subplot()   
        
    points, direction = cga2line(ga,tensor)    
    # orthogonal_shift, direction = plucker_to_line(orthogonal_shift, direction)
    for s,d in zip(t2np(points), t2np(direction)):        
        plot_line_xyz(s,d,ax, ga.p, alpha, **kargs)    
        
    if True:
        coords = t2np(points)
        for ki,xy in enumerate(coords):
            ax.text(*xy,f"{ki}") 

# eculidean
def plot_plane_xyz(d,n,ax,xx,yy, **kargs): 
    xn,yn,zn = n
    z = (d-xx*xn-yy*yn)/zn
    ax.plot_surface(xx, yy, z, alpha=0.2,**kargs)
def plot_plane_xy(d,n,ax,xx, **kargs): 
    xn,zn = n
    # z = (d-xx*xn)/zn
    z = (d-xx*xn)/zn
    ax.plot(z, xx, alpha=0.2,**kargs)

def plot_line_xyz(s,d,ax,dim, alpha = 1.0, **kargs):    
    # ic(dim,s,d) 
    _line = [(s[_]-alpha*d[_],s[_]+alpha*d[_]) for _ in range(dim)]
    ax.plot(*_line,**kargs)
    
    
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D

# Function to convert Plücker coordinates to direction vector and point on line
def plucker_to_line(p,q):
  d = np.cross(p, q)  # Direction vector
#   ic(d.shape,p.shape)
  v = (p.T / d[...,-1]).T  # Point on the line (assuming d[3] != 0)
#   v = ic(p.T / q[...,0]).T  # Point on the line (assuming d[3] != 0)
#   v = ic(p / abs(d).max(-1))  # Point on the line (assuming d[3] != 0)
# v = ic(p / d[...,-1])  # Point on the line (assuming d[3] != 0)
#   v = ic(p)  # Point on the line (assuming d[3] != 0)
  return d, v

def build_lines(d,v,alpha):
    t = alpha*np.linspace(-1, 1, 100)  # Range for parameterizing the line
    line_x = v[...,0] + t * d[...,0]
    line_y = v[...,1] + t * d[...,1]
    line_z = v[...,2] + t * d[...,2]
    return line_x, line_y, line_z 

def plot_line_xyz_(s,d,ax,dim, alpha = 1.0, **kargs):   
    d, v = plucker_to_line(-s,-d)  
    line_x, line_y, line_z = build_lines(d,v,alpha)
    # _line = [(s[_]-alpha*d[_],s[_]+alpha*d[_]) for _ in range(3)]
    # ax.plot(*_line,**kargs)
    # ax.plot_trisurf(line_x.T, line_y.T, line_z.T, cmap='viridis')  # Plot line with colormap
    ax.plot(line_x, line_y, line_z)  # Plot line with colormap

# # Define Plücker coordinates
# p = np.array([p1, p2, p3, p4, p5, p6])  # Replace with your desired values

# # Function to convert Plücker coordinates to direction vector and point on line
# def plucker_to_line(p):
#   d = np.cross(p[:3], p[3:])  # Direction vector
#   v = p[:3] / d[3]  # Point on the line (assuming d[3] != 0)
#   return d, v

# # Convert Plücker coordinates to direction vector and point
# d, v = plucker_to_line(p)

# # Create 3D plot
# fig = plt.figure()
# ax = fig.add_subplot(111, projection='3d')

# # Plot the line
# t = np.linspace(-1, 1, 100)  # Range for parameterizing the line
# line_x = v[0] + t * d[0]
# line_y = v[1] + t * d[1]
# line_z = v[2] + t * d[2]

# ax.plot_trisurf(line_x, line_y, line_z, cmap='viridis')  # Plot line with colormap

# # Set labels and title
# ax.set_xlabel('X')
# ax.set_ylabel('Y')
# ax.set_zlabel('Z')
# ax.set_title('Line in Plücker Coordinates')

# # plt.show()
    